use snafu::{ResultExt, Snafu};
use std::path::PathBuf;

#[derive(Snafu, Debug)]
pub enum Error {
  StripPrefix {
    source: std::path::StripPrefixError,
  },
  NoParent,
  RelativePath,
  #[snafu(display("Unable to canonicalize path (Path: {}, Source: {})", path.display(), source))]
  Canonicalize {
    path: PathBuf,
    source: std::io::Error,
  },
  PathOutOfBounds,
}

/// Resolves a path relative to the current path or to the base path.
///
/// Note: This function does not check that the path resolves somewhere outside
/// of the base path.
pub fn resolve_path_within_tree(
  base_path: &PathBuf,
  current_path: &PathBuf,
  path: &PathBuf,
) -> Result<PathBuf, Error> {
  let new_path = if path.is_absolute() {
    base_path.join(path.strip_prefix("/").context(StripPrefix)?)
  } else if current_path.is_dir() {
    current_path.join(path)
  } else {
    current_path
      .parent()
      .map(|x| x.join(path))
      .ok_or(Error::NoParent)?
  };

  Ok(new_path.canonicalize().context(Canonicalize {
    path: new_path.clone(),
  })?)
}
