#![feature(try_blocks, proc_macro_hygiene)]

#[macro_use]
extern crate log;

use actix::prelude::*;
use lazy_static::lazy_static;
use monitor::Monitor;
use server::Server;
use snafu::{ResultExt, Snafu};
use stats::StatsServer;
use std::sync::mpsc;
use std::{
  collections::HashSet,
  net::SocketAddr,
  sync::{Arc, RwLock},
};

pub mod bundle;
pub mod config;
pub mod files;
pub mod monitor;
pub mod server;
pub mod stats;
pub mod thread;

lazy_static! {
  static ref MONITOR: Monitor = Monitor::new();
}

#[derive(Snafu, Debug)]
pub enum Error {
  OpenConfig {
    source: config::Error,
  },
  Bind {
    address: SocketAddr,
    source: std::io::Error,
  },
  Unbundle {
    source: bundle::Error,
  },
  Serve {
    source: Box<server::Error>,
  },
  ServeStats {
    source: stats::Error,
  },
  MonitorError {
    source: monitor::Error,
  },
  RecvNotify,
}

type Result<T, E = Error> = std::result::Result<T, E>;

#[actix_rt::main]
async fn main() -> Result<()> {
  inner_main().await
}

async fn inner_main() -> Result<()> {
  MONITOR.init().context(MonitorError)?;

  pretty_env_logger::init();

  // Set up a channel for receiving thread notifications.
  let (monitor_tx, monitor_rx) = mpsc::channel();

  // Keep track of what threads have been started.
  let mut server_thread_ids = HashSet::new();
  let mut server_thread_handles = vec![];

  // Load config from default paths.
  let config = Arc::new(config::from_default_paths().context(OpenConfig)?);

  // Set up unbundler.
  let serve_dir = Arc::new(RwLock::new(None));
  let unbundler = Arc::new(bundle::Unbundler::new(config.clone(), serve_dir.clone()));

  // Set up main server.
  let server = Server::new(config.server.clone(), serve_dir);

  let (server_handle, server_join_handle, server_thread_handle) = server
    .spawn(monitor_tx.clone())
    .await
    .map_err(|err| Error::Serve {
      source: Box::new(err),
    })?;

  server_thread_ids.insert(server_join_handle.thread().id());
  server_thread_handles.push(server_thread_handle);

  // Set up optional stats server.
  let mut maybe_stats_server_handle = None;

  match &config.stats {
    Some(stats_config) => {
      let stats_server = StatsServer::new(stats_config.clone(), unbundler.clone());

      let (stats_server_handle, stats_join_handle, stats_thread_handle) = stats_server
        .spawn(monitor_tx.clone())
        .await
        .context(ServeStats)?;

      maybe_stats_server_handle = Some(stats_server_handle);
      server_thread_ids.insert(stats_join_handle.thread().id());
      server_thread_handles.push(stats_thread_handle);
    }
    None => {}
  }

  let (unbundler_join_handle, unbundler_thread_handle) =
    thread::spawn(monitor_tx.clone(), move || {
      let mut sys = System::new("unbundler");

      let result = sys
        .block_on(async move { unbundler.enter().await })
        .context(Unbundle);

      if let Err(e) = result {
        error!("Unbundler failed: {:?}", e);
      }
    });

  let (_, monitor_thread_handle) = thread::spawn(monitor_tx.clone(), move || {
    let mut watched_thread_ids = HashSet::new();

    watched_thread_ids.insert(unbundler_join_handle.thread().id());

    for server_thread_id in server_thread_ids {
      watched_thread_ids.insert(server_thread_id);
    }

    if MONITOR.watch(Some(&watched_thread_ids)).is_err() {
      error!("Failed to watch threads for panics.");
    }
  });

  // Wait for a thread to finish.
  loop {
    monitor_rx.recv().map_err(|_| Error::RecvNotify)?;

    if Ok(true) == monitor_thread_handle.has_ended() {
      info!("Stopping servers due to a panic.");

      break;
    } else if Ok(true) == unbundler_thread_handle.has_ended() {
      info!("Stopping servers due to unbundler shutdown.");

      break;
    }
  }

  // Stop server threads.
  server_handle.stop(true).await;

  if let Some(stats_server_handle) = maybe_stats_server_handle {
    stats_server_handle.stop(true).await;
  }

  Ok(())
}
